"""
Created on 2022-04-30

@author: wf
"""
import pprint
from typing import Dict, List

from ez_wikidata.wdproperty import WikidataPropertyManager
from ez_wikidata.wikidata import PropertyMapping, WdDatatype


class WikibaseQuery(object):
    """
    a Query for Wikibase
    """

    def __init__(
        self, entity: str, wpm: WikidataPropertyManager = None, debug: bool = False
    ):
        """
        Constructor

        Args:
            entity(str): the entity this query represents
            debug(bool): if True switch on debugging
        """
        self.debug = debug
        self.entity = entity
        if wpm is None:
            wpm = WikidataPropertyManager.get_instance()
        self.wpm = wpm
        self.propertiesByName = {}
        self.propertiesById = {}
        self.propertiesByVarname = {}
        self.propertiesByColumn = {}
        self.rows = []

    def get_property_mappings(self) -> List[PropertyMapping]:
        """
        Get the property mappings as PropertyMapping list

        Returns:
            List[PropertyMapping]: list of PropertyMappings
        """
        prop_maps = self.wpm.get_mappings_for_records(self.propertiesByColumn)
        return prop_maps

    def get_item_mapping(self) -> PropertyMapping:
        """
        Get the mapping that describes the wikidata entity item
        """
        return PropertyMapping.get_item_mapping(self.get_property_mappings())

    def addPropertyFromDescriptionRow(self, row):
        """
        add a property from the given row

        Args:
            row(dict): the row to add
        """
        self.rows.append(row)
        propName = row["PropertyName"]
        propId = row["PropertyId"]
        column = row["Column"]
        # properties might contain blank - replace for SPARQL variable names
        propVarname = row.get("PropVarname", propName)
        propVarname = propVarname.replace(" ", "_")
        propVarname = propVarname.replace("-", "_")
        row["PropVarname"] = propVarname
        # set the values of the lookups
        self.propertiesByName[propName] = row
        self.propertiesByColumn[column] = row
        self.propertiesById[propId] = row
        self.propertiesByVarname[propVarname] = row

    def getColumnTypeAndVarname(self, propName: str) -> (str, str, str):
        """
        get a signature tuple consisting of columnName, propertType and SPARQL variable Name for the given property Name

        Args:
            propName(str): the name of the property

        Raises:
            Exception: if property name is not known

        Returns:
            column,propType,varName tupel
        """
        if propName in self.propertiesByName:
            propRow = self.propertiesByName[propName]
            column = propRow["Column"]
            propType = propRow["Type"]
            varName = propRow["PropVarname"]
            if propType == "item" and varName in [None, ""]:
                varName = "item"
        else:
            raise Exception(
                f"unknown property name {propName} for entity {self.entity}"
            )
        return column, propType, varName

    def inFilter(
        self, values: list, propName: str = "short_name", lang: str = "en"
    ) -> str:
        """
        create a SPARQL IN filter clause

        Args:
            values(list): the list of values to filter for
            propName(str): the property name to filter with
            lang(str): the language to apply
        """
        filterClause = f"\n  FILTER(?{propName} IN("
        delim = ""
        for value in values:
            filterClause += f"{delim}\n    '{value}'@{lang}"
            delim = ","
        filterClause += "\n  ))."
        return filterClause

    def getValuesClause(
        self,
        values: list,
        propVarname: str = "short_name",
        propType: str = "text",
        lang: str = None,
        ignoreEmpty: bool = True,
        wbPrefix: str = "http://www.wikidata.org/entity/",
    ):
        """
        create a SPARQL Values clause

        Args:
            values(list): the list of values to create values for
            propVarname(str): the property variable name to assign the values for
            propType:
            lang: language of labels to query
            ignoreEmpty(bool): ignore empty values if True
            wbPrefix(str): a wikibase/wikidata prefix to be removed for items values
        Returns:
            str: the SPARQL values clause
        """
        valuesClause = f"\n  VALUES(?{propVarname}) {{"
        if lang is not None and propType == "text":
            lang = f"@{lang}"
        else:
            lang = ""
        for value in values:
            if value or not ignoreEmpty:
                if propType in ["item", "itemid", "", None]:
                    if value and value.startswith(wbPrefix):
                        value = value.replace(wbPrefix, "")
                    valuesClause += f"\n   ( wd:{value} )"
                else:
                    if isinstance(value, str):
                        # escape single quotes
                        value = value.replace("'", "\\'")
                        valuesClause += f"\n  ( '{value}'{lang} )"
                    else:
                        valuesClause += f"\n  ( {str(value)} )"
        valuesClause += "\n  }."
        return valuesClause

    def asSparql(
        self,
        filterClause: str = None,
        orderClause: str = None,
        pk: str = None,
        lang: str = "en",
    ) -> str:
        """
        get the sparqlQuery for this query optionally applying a filterClause

        Args:
            filterClause(str): a filter to be applied (if any)
            orderClause(str): an orderClause to be applied (if any)
            pk(str): primaryKey (if any)
            lang(str): the language to be used for labels
        """
        item_mapping = self.get_item_mapping()
        item_varname = item_mapping.varname
        sparql = f"""# 
# get {self.entity} records 
#  
PREFIX pq: <http://www.wikidata.org/prop/qualifier/>
PREFIX p: <http://www.wikidata.org/prop/>
PREFIX schema: <http://schema.org/>
PREFIX wd: <http://www.wikidata.org/entity/>
PREFIX wdt: <http://www.wikidata.org/prop/direct/>
PREFIX wikibase: <http://wikiba.se/ontology#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?{item_varname} ?{item_varname}Label ?{item_varname}Description
"""
        for prop_map in self.get_property_mappings():
            if prop_map.is_item_itself():
                continue
            if not prop_map.value and prop_map.varname:
                property_selections = f"\n  ?{prop_map.varname}"
                if prop_map.property_type_enum is WdDatatype.itemid:
                    # items will automatically fetch labels
                    property_selections += f" ?{prop_map.varname}Label"
                elif prop_map.property_type_enum is WdDatatype.extid:
                    # extid' will automatically fetch formatted URIs
                    property_selections += f" ?{prop_map.varname}Url"
                sparql += property_selections
        query_item_label = f"""?{item_varname} rdfs:label ?{item_varname}Label. FILTER(LANG(?{item_varname}Label) = "{lang}")"""
        query_item_desc = f"""?{item_varname} schema:description ?{item_varname}Description. FILTER(LANG(?{item_varname}Description) = "{lang}")"""
        sparql += f"""\nWHERE {{
    {query_item_label}
    OPTIONAL {{
        {query_item_desc}
    }}
"""
        for prop_map in self.get_property_mappings():
            if prop_map.propertyId in [None, ""]:
                continue
            if prop_map.value:
                # value predefined for property
                sparql += f"\n  ?{item_varname} wdt:{prop_map.propertyId} wd:{prop_map.value}."
            else:
                if prop_map.varname:
                    # primary keys are not optional
                    optional = pk is None or not prop_map.propertyName == pk
                    if optional:
                        sparql += "\n  OPTIONAL {"
                    sparql += f"\n    ?{item_varname} wdt:{prop_map.propertyId} ?{prop_map.varname}."
                    if prop_map.property_type_enum is WdDatatype.itemid:
                        # also query label of the qid with language lang
                        sparql += f"\n    ?{prop_map.varname} rdfs:label ?{prop_map.varname}Label."
                        sparql += f"""\n    FILTER(LANG(?{prop_map.varname}Label) = "{lang}")"""
                    elif prop_map.property_type_enum is WdDatatype.extid:
                        # ToDo: decision to make see https://github.com/WolfgangFahl/PyGenericSpreadSheet/issues/15
                        sparql += f"\n    wd:{prop_map.propertyId} wdt:P1630 ?{prop_map.varname}FormatterUrl."
                        sparql += f"\n    BIND(IRI(REPLACE(?{prop_map.varname}, '^(.+)$', ?{prop_map.varname}FormatterUrl)) AS ?{prop_map.varname}Url)."
                    if optional:
                        sparql += "\n  }"
        if filterClause is not None:
            sparql += f"\n{filterClause}"
        sparql += "\n}"
        if orderClause is not None:
            sparql += f"\n{orderClause}"
        return sparql

    @classmethod
    def ofMapRows(
        cls, entityMapRows: list, debug: bool = False
    ) -> Dict[str, "WikibaseQuery"]:
        """
        create a dict of wikibaseQueries from the given entityMap list of dicts

        Args:
            entityMapRows(list): a list of dict with row descriptions
            debug(bool): if True switch on debugging
        """
        queries = {}
        entityMapDict = {}
        for row in entityMapRows:
            if "Entity" in row:
                entity = row["Entity"]
                if not entity in entityMapDict:
                    entityMapDict[entity] = {}
                entityRows = entityMapDict[entity]
                if "PropertyName" in row:
                    propertyName = row["PropertyName"]
                    entityRows[propertyName] = row
        if debug:
            pprint.pprint(entityMapDict)
        for entity in entityMapDict:
            wbQuery = WikibaseQuery.ofEntityMap(entity, entityMapDict[entity])
            queries[entity] = wbQuery
        return queries

    @classmethod
    def ofEntityMap(cls, entity: str, entityMap: dict) -> "WikibaseQuery":
        """
        create a WikibaseQuery for the given entity and entityMap

        Args:
            entity(str): the entity name
            entityMap(dict): the entity property descriptions
        Returns:
            WikibaseQuery
        """
        wbQuery = WikibaseQuery(entity)
        for row in entityMap.values():
            wbQuery.addPropertyFromDescriptionRow(row)
        return wbQuery



"""
Created on 2022-04-18

@author: wf
"""
import datetime
import json
import os
import re
import textwrap
import traceback
import typing
from dataclasses import dataclass, field
from itertools import groupby
from pathlib import Path
from typing import Dict, List, Optional, Union

import dateutil.parser
from lodstorage.sparql import SPARQL
from wikibaseintegrator import WikibaseIntegrator, wbi_login
from wikibaseintegrator.datatypes import (
    URL,
    BaseDataType,
    ExternalID,
    Item,
    MonolingualText,
    String,
    Time,
)
from wikibaseintegrator.entities import ItemEntity
from wikibaseintegrator.models import Claim, Reference, Snak
from wikibaseintegrator.wbi_config import config as wbi_config
from wikibaseintegrator.wbi_enums import WikibaseDatePrecision, WikibaseRank

from ez_wikidata.prefixes import Prefixes
from ez_wikidata.version import Version
from ez_wikidata.wdproperty import (
    PropertyMapping,
    PropertyMappings,
    Variable,
    WdDatatype,
    WikidataPropertyManager,
)


@dataclass
class WikidataResult:
    """
    a class for handling a wikidata result
    """

    item: Optional[ItemEntity] = None
    errors: Dict[str, Exception] = field(default_factory=dict)
    qid: Optional[str] = None
    msg: Optional[str] = None
    debug: Optional[bool] = False

    def __post_init__(self):
        # If qid is not provided, derive it from item
        if self.qid is None and self.item:
            self.qid = self.item.id

    @property
    def pretty_item_json(self, indent: int = 2) -> str:
        """Returns a pretty-printed JSON string of the item."""
        if self.item:
            item_dict = (
                self.item.get_json()
            )  # Assuming get_json() returns a JSON string representation of the item
            pretty_json = json.dumps(item_dict, indent=indent)
        else:
            pretty_json = self.qid
        return pretty_json


class Wikidata:
    """
    wikidata access

    see http://learningwikibase.com/data-import/
    """

    TEST_WD_URL = "https://test.wikidata.org"
    WD_URL = "https://www.wikidata.org"

    def __init__(
        self,
        baseurl: str = None,
        wpm: WikidataPropertyManager = None,
        debug: bool = False,
    ):
        """
        Constructor

        Args:
            baseurl(str): the baseurl of the wikibase to use
            debug(bool): if True output debug information
            wpm(WikidataPropertymanager):
        """
        if baseurl is None:
            baseurl = self.WD_URL
        self.baseurl = baseurl
        self.debug = debug
        self.apiurl = f"{self.baseurl}/w/api.php"
        self.login = None
        self.user = None
        self._wbi = None
        if wpm is None:
            wpm = WikidataPropertyManager.get_instance()
        self.wpm = wpm

    @property
    def wbi(self) -> WikibaseIntegrator:
        """
        WikibaseIntegrator
        """
        if self._wbi is None or (self.login is not None and self._wbi.login is None):
            wbi_config[
                "USER_AGENT"
            ] = f"{Version.name}/{Version.version} (https://www.wikidata.org/wiki/User:{self.user})"
            wbi_config["MEDIAWIKI_API_URL"] = self.apiurl
            self._wbi = WikibaseIntegrator(login=self.login)
        return self._wbi

    @wbi.setter
    def wbi(self, wbi: typing.Union[WikibaseIntegrator, None]):
        """
        set the WikibaseIntegrator
        """
        self._wbi = wbi

    def getCredentials(self) -> (str, str):
        """
        get my credentials https://test.wikidata.org/wiki/Property:P370

        from the wd npm command line tool

        Throws:
            Exception: if no credentials are available for the baseurl

        Returns:
            (username, password) of the account assigned to the baseurl
        """
        user = None
        pwd = None
        home = str(Path.home())
        configFilePath = f"{home}/.config/wikibase-cli/config.json"
        if os.path.isfile(configFilePath):
            with open(configFilePath, mode="r") as f:
                wikibaseConfigJson = json.load(f)
                credentials = wikibaseConfigJson["credentials"]
                credentialRecord = credentials.get(self.baseurl, None)
                if (
                    self.baseurl == self.TEST_WD_URL
                    and self.baseurl not in credentials
                    and self.WD_URL in credentials
                ):
                    credentialRecord = credentials.get(self.WD_URL)
                if credentialRecord is None:
                    raise Exception(f"no credentials available for {self.baseurl}")
                user = credentialRecord["username"]
                pwd = credentialRecord["password"]
        return user, pwd

    def loginWithCredentials(self, user: str = None, pwd: str = None):
        """
        login using the given credentials or credentials
        retrieved via self.getCredentials

        Args:
            user(str): the username
            pwd(str): the password
        """
        if user is None:
            user, pwd = self.getCredentials()

        if user is not None:
            self.login = wbi_login.Login(
                user=user, password=pwd, mediawiki_api_url=self.apiurl
            )
            if self.login:
                self.user = user

    def logout(self):
        """
        log the user out again
        """
        self.user = None
        self.login = None
        self.wbi = None

    def getItemByName(
        self, itemName: str, itemType: str, lang: str = "en"
    ) -> typing.Optional[str]:
        """
        get an item by Name
        ToDo: Needs to be reworked as always WDQS is used as endpoint even if a different one is defined
        Args:
            itemName(str): the item to look for
            itemType(str): the type of the item
            lang(str): the language of the itemName
        """
        itemLabel = f'"{itemName}"@{lang}'
        sparqlQuery = """PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX wdt: <http://www.wikidata.org/prop/direct/>
            PREFIX wd: <http://www.wikidata.org/entity/>
            
            SELECT ?item ?itemLabel
            WHERE {
              {
                ?item wdt:P31|wdt:P31/wdt:P279 wd:%s.
                ?item rdfs:label ?itemLabel.
                # short name
                BIND(%s as ?shortNameLabel )
                ?item wdt:P1813 ?shortNameLabel
                FILTER(LANG(?itemLabel)= "%s" )
              } UNION {
                ?item wdt:P31|wdt:P31/wdt:P279 wd:%s.
                BIND(%s as ?itemLabel )
                ?item rdfs:label ?itemLabel.
              }
            }""" % (
            itemType,
            itemLabel,
            lang,
            itemType,
            itemLabel,
        )
        endpointUrl = "https://query.wikidata.org/sparql"
        sparql = SPARQL(endpointUrl)
        itemRows = sparql.queryAsListOfDicts(sparqlQuery)
        item = None
        if len(itemRows) > 0:
            item = itemRows[0]["item"].replace("http://www.wikidata.org/entity/", "")
        return item

    def addDict(
        self,
        row: dict,
        mapDict: dict,
        itemId: Union[str, None] = None,
        lang: str = "en",
        write: bool = False,
        ignoreErrors: bool = False,
    ) -> WikidataResult:
        """
        add the given row mapping with the given map Dict

        Args:
            row(dict): the data row to add
            mapDict(dict): the mapping dictionary to use
            itemId: wikidata id of the item the data should be added to. If None a new item is created unless item id is provided in the record
            lang(str): the language for lookups
            write(bool): if True do actually write
            ignoreErrors(bool): if True ignore errors

        Returns:
            WikiDataResult: the result of the operation
        """
        mappings = self.wpm.get_mappings_for_records(mapDict)
        return self.add_record(
            row,
            mappings,
            item_id=itemId,
            lang=lang,
            write=write,
            ignore_errors=ignoreErrors,
        )

    def get_record(
        self,
        item_id: str,
        property_mappings: Union[
            List[str], List["PropertyMapping"], typing.Dict[str, dict]
        ],
        include_label: bool = True,
        include_description: bool = True,
        label_for_qids: bool = False,
    ) -> dict:
        """
        Get the properties form the given item
        Args:
            item_id: id of the item to get the data from
            property_mappings: list of property values to extract
            include_label:
            include_description:
            label_for_qids: If True fetch the label for a linked Qid
        Returns:
            dict with the property values
        """
        item = self.wbi.item.get(item_id)
        lang = "en"
        if isinstance(property_mappings, dict):
            property_mappings = PropertyMapping.from_records(property_mappings)
        record = dict()
        if include_label and item.labels.get(lang) is not None:
            record["label"] = item.labels.get(lang).value
        if include_description and item.descriptions.get(lang) is not None:
            record["description"] = item.descriptions.get(lang).value
        qualifier_lookup = PropertyMapping.get_qualifier_lookup(property_mappings)
        pms = []
        for pm in property_mappings:
            if not isinstance(pm, PropertyMapping) or not pm.is_qualifier():
                pms.append(pm)
        for prop in pms:
            prop_id = prop
            if isinstance(prop, PropertyMapping):
                prop_id = prop.propertyId
            statements = self._get_statements_by_pid(item, prop_id)
            prop_label = prop_id
            if isinstance(prop, PropertyMapping):
                prop_label = prop.column
            values = []
            for statement in statements:
                value = self._get_statement_value(statement)
                if label_for_qids:
                    if (
                        prop.valueLookupType is not None
                        and statement.mainsnak.datatype == "wikibase-item"
                    ):
                        label = self.get_item_label(value, lang)
                        if label is not None:
                            value = label
                values.append(value)
                if (
                    isinstance(prop, PropertyMapping)
                    and prop.column in qualifier_lookup
                ):
                    for qualifier_pm in qualifier_lookup[prop.column]:
                        if qualifier_pm.propertyId in statement.qualifiers.qualifiers:
                            qualifier_statements = statement.qualifiers.get(
                                qualifier_pm.propertyId
                            )
                        else:
                            qualifier_statements = []
                        qualifier_values = []
                        for qualifier_statement in qualifier_statements:
                            qualifier_values.append(
                                self._get_statement_value(qualifier_statement)
                            )
                        record[qualifier_pm.column] = (
                            qualifier_values[0]
                            if len(qualifier_values) == 1
                            else qualifier_values
                        )
            if len(values) == 1:
                record[prop_label] = values[0]
            elif values == []:
                record[prop_label] = None
            else:
                record[prop_label] = values
        return record

    def get_item_label(self, item_id: str, lang: str = None) -> typing.Union[str, None]:
        """
        Get the label for the given item id
        Args:
            item_id: id of the item
            lang: label language to return. Default is "en"

        Returns:
            str: label of the item
            None: If the label can not be determined or the item_id is None or can not be found
        """
        if lang is None:
            lang = "en"
        label = None
        if item_id is not None:
            linked_item = self.wbi.item.get(item_id)
            linked_item_label = linked_item.labels.get(lang)
            if linked_item_label is not None:
                label = linked_item_label.value
        return label

    def _get_statements_by_pid(self, item: ItemEntity, pid: str) -> List[Item]:
        """
        Get the property statements of the item for the given Pid.
        if ranking is established between the statements return only the highest rank
        Args:
            item: item to get the statements from
            pid: property id
        Returns:
            list: list of the property statements
        """
        if pid in item.claims:
            statements = item.claims.get(pid)
        else:
            statements = []
        if len(statements) > 1:
            ordered_stats = {
                k: list(g) for k, g in groupby(statements, lambda x: x.rank)
            }
            rank_by_preference = [
                WikibaseRank.PREFERRED,
                WikibaseRank.NORMAL,
                WikibaseRank.DEPRECATED,
            ]
            for rank in rank_by_preference:
                if rank in ordered_stats:
                    statements = ordered_stats[rank]
                    break
        return statements

    def _get_statement_value(self, statement: Union[Claim, Snak]) -> typing.Any:
        """
        Get the raw value of the statement without the metadata
        Args:
            statement: statement to extract the value from

        Returns:
            raw value of the statement
        """
        value = None
        snak = statement
        if isinstance(statement, Claim):
            snak = statement.mainsnak
        raw_value = snak.datavalue.get("value")
        datatype = snak.datatype
        if datatype == "wikibase-item":
            value = raw_value.get("id", None)
        elif datatype == "monolingualtext":
            value = raw_value.get("text")
        elif datatype == "string":
            value = raw_value
        elif datatype == "url":
            value = raw_value
        elif datatype == "time":
            value = dateutil.parser.parse(raw_value.get("time")[1:])
            precision = raw_value.get("precision")
            if precision == 11:
                value = value.date()
            elif precision == 9:
                value = value.year
        elif datatype == "external-id":
            value = raw_value
        else:
            pass
        return value

    def add_record(
        self,
        record: dict,
        property_mappings: List["PropertyMapping"],
        item_id: Union[str, None] = None,
        lang: str = "en",
        write: bool = False,
        ignore_errors: bool = False,
        summary: str = None,
        reference: Reference = None,
    ) -> WikidataResult:
        """
        add the given row mapping with the given map Dict

        Args:
            record(dict): the data row to add
            property_mappings(list): the mapping dictionary to use
            item_id: wikidata id of the item the data should be added to. If None a new item is created unless item id is provided in the record
            lang(str): the language for lookups
            write(bool): if True do actually write
            ignore_errors(bool): if True ignore errors
            summary: summary of the item edits
            reference: reference to add to all claims

        Returns:
            (qId, errors): the wikidata item create (if any) and a dict of errors
        """
        claims = []
        errors = dict()
        qualifier_lookup = PropertyMapping.get_qualifier_lookup(property_mappings)
        # check if there is a existing Q-Item identifier in the record
        item_mapping = PropertyMapping.get_item_mapping(property_mappings)
        if item_mapping is not None:
            if item_id is None:
                item_id = record.get(item_mapping.column, None)
        # get the relevant properties
        properties = []
        for pm in property_mappings:
            if not pm.is_qualifier() and not pm.is_item_itself():
                properties.append(pm)
            else:
                # breakpoint to debug ignored properties
                pass

        for prop in properties:
            qualifier_mappings = qualifier_lookup.get(prop.column, None)
            prop_claims, claim_errors = self._get_statement_for_property(
                record, prop, qualifier_mappings, reference, lang
            )
            # merge error dicts to one dict
            errors = {**errors, **claim_errors}
            claims.extend(prop_claims)
        label = self.sanitize_label(record.get("label", None))
        description = record.get("description", None)
        # handle get or create case
        item = self.get_or_create_item(item_id)
        item.add_claims(claims)
        if label:
            item.labels.set(language=lang, value=label)
        if description:
            item.descriptions.set(language=lang, value=description)
        if write:
            if len(errors) == 0 or ignore_errors:
                try:
                    item = item.write(summary=summary)
                except Exception as ex:
                    errors["write failed"] = ex
        result = WikidataResult(item=item, errors=errors, debug=self.debug)
        return result

    def _get_statement_for_property(
        self,
        record: dict,
        prop_mapping: "PropertyMapping",
        qualifier_mappings: Union[List["PropertyMapping"], None],
        reference: Reference,
        lang: str,
    ) -> (List[Claim], dict):
        """
        Get the claims that can be derived from the given property mapping and record.
        Generates a statement with its qualifiers and reference from the given record and mapping.
        If the record value of the property is a list multiple claims are generated

        Args:
            record: data record
            prop_mapping: property definition for the claims that should be generated from the given record
            qualifier_mappings: descriptions of the qualifiers of the property
            reference: reference of the statement
            lang: language to use

        Returns:
            list of statements
        """
        claims = []
        value = self.get_prop_value(record, prop_mapping, lang)
        values = value if isinstance(value, list) else [value]
        errors = dict()
        for value in values:
            statement = None
            try:
                statement = self.convert_to_claim(value=value, pm=prop_mapping)
            except Exception as ex:
                errors[prop_mapping.column] = ex
                if self.debug:
                    print(traceback.format_exc())
            if statement is not None:
                # add reference
                if reference is not None:
                    statement.references.add(reference)
                # add qualifier
                if qualifier_mappings is not None:
                    qualifier_errors = self._add_qualifier_to_statement(
                        record, statement, qualifier_mappings, lang
                    )
                    # merge error dicts to one dict
                    errors = {**errors, **qualifier_errors}
            if statement is not None:
                claims.append(statement)
        return claims, errors

    def _add_qualifier_to_statement(
        self,
        record: dict,
        statement: Claim,
        qualifier_mappings: List["PropertyMapping"],
        lang: str,
    ) -> dict:
        """
        add the qualifiers to the given statement
        Args:
            record:
            statement: add qualifiers to this statement
            qualifier_mappings: list of PropertyMappings of the qualifiers

        Returns:
            dict of occurred errors with the qualifier column as key. If no error occurs an empty dict is returned
        """
        errors = dict()
        for qualifier_pm in qualifier_mappings:
            qualifier_value = self.get_prop_value(record, qualifier_pm, lang)
            if qualifier_value is None:
                continue
            else:
                try:
                    qualifier = self.convert_to_claim(qualifier_value, qualifier_pm)
                    statement.qualifiers.add(qualifier)
                except Exception as ex:
                    errors[qualifier_pm.column] = ex
                    if self.debug:
                        print(traceback.format_exc())
        return errors

    def get_or_create_item(self, item_id: typing.Union[str, None]) -> ItemEntity:
        """
        Get or create the requested wikidata item
        Args:
            item_id: item to retrieve if None create a new item
        """
        if item_id is None or isinstance(item_id, str) and item_id.strip() == "":
            item = self.wbi.item.new()
        else:
            item = self.wbi.item.get(item_id)
        return item

    def get_prop_value(
        self, record: dict, pm: "PropertyMapping", lang: str
    ) -> typing.Any:
        """
        Retrieve the property value from the record and prepare the value if necessary
        Args:
            record: record containing the property data
            pm: property mapping
            lang: language to use

        Returns:
            value of the property from the record
        """
        value = record.get(pm.column, None)
        if value is None:
            value = pm.value
        if value and pm.valueLookupType and not self.is_wikidata_item_id(value):
            # find the wikidata item id of value
            value = self.getItemByName(value, pm.valueLookupType, lang)
        if value and isinstance(value, str):
            value = value.strip()
        return value

    def convert_to_claim(
        self, value, pm: "PropertyMapping"
    ) -> Union[BaseDataType, None]:
        """
        Convert the given value to a corresponding wikidata statement
        Args:
            value: value of the statement
            pm: information about the property statement ot generate

        Raises:
            Exception: if property datatype is unknown or not supported

        Returns:
            BaseDataType
        """
        if value is None or value == "":
            return None
        if pm.property_type_enum is None:
            pm.property_type_enum = self.get_wddatatype_of_property(pm.propertyId)
        if pm.property_type_enum is WdDatatype.year:
            yearString = f"+{value}-01-01T00:00:00Z"
            statement = Time(
                yearString, prop_nr=pm.propertyId, precision=WikibaseDatePrecision.YEAR
            )
        elif pm.property_type_enum is WdDatatype.date:
            statement = self.get_date_claim(value, pm.propertyId)
        elif pm.property_type_enum is WdDatatype.extid:
            statement = ExternalID(value=value, prop_nr=pm.propertyId)
        elif pm.property_type_enum is WdDatatype.string:
            statement = String(value=str(value), prop_nr=pm.propertyId)
        elif pm.property_type_enum is WdDatatype.text:
            statement = MonolingualText(text=str(value), prop_nr=pm.propertyId)
        elif pm.property_type_enum is WdDatatype.url:
            statement = URL(value=value, prop_nr=pm.propertyId)
        elif pm.property_type_enum is WdDatatype.itemid:
            statement = Item(value=value, prop_nr=pm.propertyId)
        else:
            raise Exception(
                f"({pm.property_type_enum}) unknown or not supported datatype"
            )
        return statement

    @staticmethod
    def get_date_claim(
        date: Union[str, datetime.date, datetime.datetime], prop_nr: Union[str, int]
    ) -> Claim:
        """
        Get the data statement for the given date and property id
        Args:
            date: date value
            prop_nr: id of the property

        Returns:
            statement of the given property number with the given value
        """
        if isinstance(date, datetime.date):
            date_value = datetime.datetime.combine(date, datetime.time())
        elif isinstance(date, datetime.datetime):
            date_value = date
        elif isinstance(date, str):
            date_value = dateutil.parser.parse(date)
        else:
            raise Exception(f"Value '{date}' can not be parsed to date")
        iso_date = date_value.isoformat()
        date_string = f"+{iso_date}Z"
        statement = Time(
            date_string, prop_nr=prop_nr, precision=WikibaseDatePrecision.DAY
        )
        return statement

    @staticmethod
    def is_wikidata_item_id(value: str) -> bool:
        """
        Returns true if the given value is a wikidata item id
        """
        return bool(re.fullmatch(r"Q[0-9]+", value))

    @staticmethod
    def is_wikidata_property_id(value: str) -> bool:
        """
        Returns true if the given value is a wikidata property id
        """
        return bool(re.fullmatch(r"P[0-9]+", value))

    @staticmethod
    def sanitize_label(label: str, limit: int = None, postfix: str = None) -> str:
        """
        sanitize given label by ensuring it is not too long
        Args:
            label: label to sanitize
            limit: max length of the label

        Returns:
            sanitized label
        """
        if limit is None:
            limit = 250
        if postfix is None:
            postfix = "..."
        if label is not None and len(label) > limit:
            label = label[: limit - len(postfix)] + postfix
        return label

    @classmethod
    def get_datatype_of_property(cls, property_id: Union[str, int]) -> Union[str, None]:
        """
        Get the datatype of the given property
        Args:
            property_id: id of the property e.g. P31 or 31

        Returns:
            datatype of the property of None if no datatype is defined
        """
        if isinstance(property_id, int) or not property_id.startswith("P"):
            property_id = f"P{property_id}"
        query = """
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX wdt: <http://www.wikidata.org/prop/direct/>
            PREFIX wd: <http://www.wikidata.org/entity/>
            PREFIX wikibase: <http://wikiba.se/ontology#>
            
            SELECT Distinct ?o
            WHERE {
              wd:%s wikibase:propertyType ?o.
            }
        """ % (
            property_id
        )
        endpointUrl = "https://query.wikidata.org/sparql"
        sparql = SPARQL(endpointUrl)
        itemRows = sparql.queryAsListOfDicts(query)
        wikibase_prefix = "http://wikiba.se/ontology#"
        types = []
        for record in itemRows:
            types.append(record.get("o")[len(wikibase_prefix) :])
        if len(types) > 1:
            print("Property has more than one type! please check")
        elif len(types) == 0:
            print("Property has no defined datatype! please check")
            return None
        return types[0]

    @classmethod
    def get_wddatatype_of_property(cls, property_id: Union[str, int]) -> "WdDatatype":
        """
        Get the datatype of the given property
        Args:
            property_id: id of the property e.g. P31 or 31

        Returns:
            WdDatatype of the property of None if no datatype is defined
        """
        property_type = cls.get_datatype_of_property(property_id)
        return WdDatatype.get_by_wikibase(property_type)

    def normalize_records(self, record: dict, prop_map: typing.List["PropertyMapping"]):
        """
        Normalize given record by converting Qids to WikidataItem objects (lookup label) and find out Qid if label given
        based on the given prop_map
        """
        itemid_props = [
            p for p in prop_map if p.property_type_enum is WdDatatype.itemid
        ]
        for p in itemid_props:
            if p.column is None or p.column == "":
                continue
            value = record.get(p.column, None)
            if value is None and p.value is not None:
                value = p.value
            if isinstance(value, list):
                wd_item = [self.get_wikidata_item(v, p.valueLookupType) for v in value]
            else:
                wd_item = self.get_wikidata_item(value, p.valueLookupType)
            record[p.column] = wd_item
        return record

    def get_wikidata_item(
        self, qid_or_label: str, item_type_qid: str = None
    ) -> typing.Optional["WikidataItem"]:
        """
        Get WikidataItem for given label or Qid

        Args:
            qid_or_label: label or Qid of a item

        Returns:
            WikidataItem
        """
        item = None
        if qid_or_label is not None:
            if self.is_wikidata_item_id(qid_or_label):
                # lookup label
                qid = qid_or_label
                label = self.get_item_label(qid)
            else:
                # lookup label
                label = qid_or_label
                qid = self.getItemByName(label, item_type_qid)
            if qid is not None:
                item = WikidataItem(qid, label)
        return item


@dataclass
class WikidataItem:
    qid: str
    label: str
    lang: str = "en"
    sparql: Optional[SPARQL] = None
    debug: bool = False
    label: str = field(init=False, default=None)
    description: str = field(init=False, default=None)
    url: str = field(init=False)

    def __eq__(self, other) -> bool:
        """
        WikidataItems are equal if the qid is equal
        """
        same = isinstance(other, WikidataItem) and self.qid == getattr(
            other, "qid", None
        )
        return same

    def __post_init__(self):
        """
        handle the construction
        """
        if not self.qid:
            self.qid = None
            return
        self.url = f"https://www.wikidata.org/wiki/{self.qid}"
        # numeric qid
        self.qnumber = int(self.qid[1:])
        self.url = f"https://www.wikidata.org/wiki/{self.qid}"
        if self.sparql is not None:
            self.qlabel, self.description = WikidataItem.getLabelAndDescription(
                self.sparql, self.qid, self.lang, debug=self.debug
            )
            self.varname = Variable.validVarName(self.qlabel)
            self.itemVarname = f"{self.varname}Item"
            self.labelVarname = f"{self.varname}"

    def __str__(self):
        return self.asText(long=False)

    def asText(self, long: bool = True, wrapAt: int = 0):
        """
        returns my content as a text representation

        Args:
            long(bool): True if a long format including url is wished
            wrapAt(int): wrap long lines at the given width (if >0)

        Returns:
            str: a text representation of my content
        """
        text = self.qid or "❓"
        if hasattr(self, "qlabel"):
            text = f"{self.qlabel} ({self.qid})"
        if hasattr(self, "description"):
            desc = self.description
            if wrapAt > 0:
                desc = textwrap.fill(desc, width=wrapAt)
            text += f"☞{desc}"
        if long and hasattr(self, "url"):
            text += f"→ {self.url}"
        return text

    @classmethod
    def getLabelAndDescription(
        cls, sparql: SPARQL, itemId: str, lang: str = "en", debug: bool = False
    ):
        """
        get  the label for the given item and language

        Args:
            itemId(str): the wikidata Q/P id
            lang(str): the language of the label
            debug(bool): if True output debug information

        Returns:
            (str,str): the label and description as a tuple
        """
        query = f"""# get the label for the given item
{Prefixes.getPrefixes(["rdfs","wd","schema"])}        
SELECT ?itemLabel ?itemDescription
WHERE
{{
  VALUES ?item {{
    wd:{itemId}
  }}
  ?item rdfs:label ?itemLabel.
  FILTER (LANG(?itemLabel) = "{lang}").
  ?item schema:description ?itemDescription.
  FILTER(LANG(?itemDescription) = "{lang}")
}}"""
        try:
            if debug:
                msg = f"getLabelAndDescription for wikidata Item {itemId} with query:\n{query}"
                print(msg)
            labelAndDescription = sparql.getValues(
                query, ["itemLabel", "itemDescription"]
            )
        except Exception as ex:
            msg = f"getLabelAndDescription failed for wikidata Item {itemId}:{str(ex)}"
            if debug:
                print(msg)
            raise Exception(msg)
        return labelAndDescription

    @classmethod
    def getItemsByLabel(
        cls, sparql: SPARQL, itemLabel: str, lang: str = "en", debug: bool = False
    ) -> list:
        """
        get a Wikidata items by the given label

        Args:
            sparql(SPARQL): the SPARQL endpoint to use
            itemLabel(str): the label of the items
            lang(str): the language of the label
            debug(bool): if True show debugging information

        Returns:
            a list of potential items
        """
        valuesClause = f'   "{itemLabel}"@{lang}\n'
        query = f"""# get the items that have the given label in the given language
# e.g. we'll find human=Q5 as the oldest type for the label "human" first
# and then the newer ones such as "race in Warcraft"
{Prefixes.getPrefixes(["rdfs","schema","xsd"])}
SELECT 
  #?itemId 
  ?item 
  ?itemLabel 
  ?itemDescription
WHERE {{ 
  VALUES ?itemLabel {{
    {valuesClause}
  }}
  #BIND (xsd:integer(SUBSTR(STR(?item),33)) AS ?itemId)
  ?item rdfs:label ?itemLabel. 
  ?item schema:description ?itemDescription.
  FILTER(LANG(?itemDescription)="{lang}")
}} 
#ORDER BY ?itemId"""
        qLod = sparql.queryAsListOfDicts(query)
        items = []
        for record in qLod:
            url = record["item"]
            qid = re.sub(r"http://www.wikidata.org/entity/(.*)", r"\1", url)
            item = WikidataItem(qid, debug=debug)
            item.url = url
            item.qlabel = record["itemLabel"]
            item.varname = Variable.validVarName(item.qlabel)
            item.description = record["itemDescription"]
            items.append(item)
        sortedItems = sorted(items, key=lambda item: item.qnumber)
        return sortedItems


class UrlReference(Reference):
    """
    Reference consisting of
        reference URL (P854)
        retrieved (P813)
    """

    def __init__(
        self, url, date: Union[str, datetime.date, datetime.datetime, None] = None
    ):
        """
        constructor
        Args:
            url: reference URL
            date: retrieved at
        """
        super().__init__()
        self.url = url
        if date is None:
            date = datetime.date.today()
        self.date = date
        self.add(URL(value=self.url, prop_nr="P854"))
        self.add(Wikidata.get_date_claim(date, prop_nr="P813"))



"""
Created on 02.03.2024-03-02

@author: wf
"""
import os
import re
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from lodstorage.profiler import Profiler
from lodstorage.sparql import SPARQL
from lodstorage.sql import SQLDB
from lodstorage.yamlable import lod_storable

from ez_wikidata.prefixes import Prefixes


class WdDatatype(Enum):
    """
    Supported Wikidata datatypes, sorted by frequency and including special cases.
    """

    extid = auto()  # ExternalId: 8645 occurrences
    itemid = auto()  # WikibaseItem: 1634 occurrences
    quantity = auto()  # Quantity: 652 occurrences
    string = auto()  # String: 329 occurrences
    url = auto()  # Url: 107 occurrences
    commons_media = auto()  # CommonsMedia: 79 occurrences
    time = auto()  # Time: 66 occurrences
    text = auto()  # Monolingualtext: 62 occurrences
    math = auto()  # Math: 36 occurrences
    wikibase_property = auto()  # WikibaseProperty: 21 occurrences
    wikibase_sense = auto()  # WikibaseSense: 19 occurrences
    wikibase_lexeme = auto()  # WikibaseLexeme: 17 occurrences
    globe_coordinate = auto()  # GlobeCoordinate: 11 occurrences
    wikibase_form = auto()  # WikibaseForm: 9 occurrences
    musical_notation = auto()  # MusicalNotation: 6 occurrences
    tabular_data = auto()  # TabularData: 6 occurrences
    geoshape = auto()  # GeoShape: 3 occurrences
    # Special cases:
    item = auto()  # Item: Special case
    year = auto()  # Year: Special case
    date = auto()  # Date: Special case

    @classmethod
    def from_wb_type_name(cls, wb_type_name: str) -> "WdDatatype":
        """
        convert a wikibase type name to a WdDatatype

        Args:
            wb_type_name(str): the string name of the wikibase type (with or without wikibase ontology prefix)
        """
        type_map = {
            "ExternalId": cls.extid,
            "WikibaseItem": cls.itemid,
            "Quantity": cls.quantity,
            "String": cls.string,
            "Url": cls.url,
            "CommonsMedia": cls.commons_media,
            "Time": cls.time,
            "Monolingualtext": cls.text,
            "Math": cls.math,
            "WikibaseProperty": cls.wikibase_property,
            "WikibaseSense": cls.wikibase_sense,
            "WikibaseLexeme": cls.wikibase_lexeme,
            "GlobeCoordinate": cls.globe_coordinate,
            "WikibaseForm": cls.wikibase_form,
            "MusicalNotation": cls.musical_notation,
            "TabularData": cls.tabular_data,
            "GeoShape": cls.geoshape,
        }
        wb_type_name = wb_type_name.replace("http://wikiba.se/ontology#", "")
        wd_type = type_map.get(wb_type_name, WdDatatype.string)
        return wd_type

    @classmethod
    def _missing_(cls, _value):
        """
        default datatype
        """
        return cls.text

    @classmethod
    def get_by_wikibase(cls, property_type: str) -> Union["WdDatatype", None]:
        """
        Get WdDatatype by the corresponding wikibase datatype
        Args:
            property_type: wikibase name of the type

        Returns:
            WdDatatype
        """
        wikibase_map = {
            "WikibaseItem": cls.itemid,
            "Time": cls.date,
            "Monolingualtext": cls.text,
            "String": cls.string,
            "ExternalId": cls.extid,
            "Url": cls.url,
        }
        return wikibase_map.get(property_type, None)


class Variable:
    """
    Variable e.g. name handling
    """

    @classmethod
    def validVarName(cls, varStr: str) -> str:
        """
        convert the given potential variable name string to a valid
        variable name

        see https://stackoverflow.com/a/3305731/1497139

        Args:
            varStr(str): the string to convert

        Returns:
            str: a valid variable name
        """
        return re.sub("\W|^(?=\d)", "_", varStr)


@lod_storable
class WikidataProperty:
    """
    Represents a Wikidata Property.
    """
    id: str # the id of the property - pid + lang
    pid: str  # The property ID
    lang: str
    plabel: str  # the label of the property
    description: str  # Description of the property
    type_name: str  # the type name
    reverse: bool = False  # Indicates if the property is used in reverse direction
    # Variables initialized in __post_init__
    # varname: str = field(init=False)
    # valueVarname: str = field(init=False)
    # labelVarname: str = field(init=False)
    # ptype: WdDatatype = field(init=False)

    def __post_init__(self):
        """
        creates and modify calculated fields
        """
        # not needed any more but does not hurt
        self.pid = self.pid.replace("http://www.wikidata.org/entity/", "")
        self.url = f"https://www.wikidata.org/wiki/Property:{self.pid}"
        self.ptype = WdDatatype.from_wb_type_name(self.type_name)
        self.varname = Variable.validVarName(self.plabel)
        self.valueVarname = (
            f"{self.varname}Item" if "WikibaseItem" in self.type_name else self.varname
        )
        self.labelVarname = self.varname

    def getPredicate(self):
        """
        get me as a Predicate
        """
        reverseToken = "^" if self.reverse else ""
        plabel = f"{reverseToken}wdt:{self.pid}"
        return plabel

    def __str__(self):
        text = self.pid
        if hasattr(self, "plabel"):
            text = f"{self.plabel} ({self.pid})"
        return text


class WikidataPropertyManager:
    """
    handle Wikidata Properties
    """

    def __init__(
        self,
        endpoint_url: str = "https://qlever.cs.uni-freiburg.de/api/wikidata",
        langs: List[str] = ["en", "zh", "hi", "de", "fr", "ar", "es", "bn", "ru"],
        with_load: bool = True,
        profile: bool = True,
        debug: bool = False,
    ):
        """
        initialize the lookups
        """
        if not "en" in langs:
            raise ValueError(f"en is mandatory in langs -{langs}")
        self.langs = langs
        self.debug = debug
        self.profile = profile
        self.sparql = SPARQL(endpoint_url, debug=self.debug)
        self.sql_db_path = WikidataPropertyManager.get_cache_path()
        self.sql_db = SQLDB(self.sql_db_path)
        self.sparql_query = self.get_query_for_langs(langs)
        self.props = []
        self.props_by_id = {}
        self.props_by_lang = {}
        self.loaded=False
        if with_load:
            self.load()

    def load_from_sparql(self):
        """
        get my list of dicts from sparql
        """
        profiler = Profiler(f"getting wikidata properties for {len(self.langs)} languages via SPARQL", profile=self.profile)
        self.lod = self.sparql.queryAsListOfDicts(self.sparql_query)
        profiler.time()

    def store(self):
        """
        store my list of dicts
        """
        profiler = Profiler(f"caching wikidata properties to SQL", profile=self.profile)
        self.entity_info = self.sql_db.createTable(
            listOfRecords=self.lod,
            entityName="wd_properties",
            primaryKey="id",
            withCreate=True,
            withDrop=True,
            sampleRecordCount=100,
        )
        self.sql_db.store(
            listOfRecords=self.lod,
            entityInfo=self.entity_info,
            executeMany=True,
            fixNone=True,
        )
        profiler.time()

    def load_from_sql(self):
        """
        load from SQL
        """
        profiler = Profiler(
            f"loading wikidata properties from SQL", profile=self.profile
        )
        sql_query = "SELECT * FROM wd_properties"
        self.lod = self.sql_db.query(sql_query)
        profiler.time()

    def load(self):
        """
        load the properties
        """
        if self.loaded:
            return
        if os.path.isfile(self.sql_db_path) and os.stat(self.sql_db_path).st_size > 0:
            self.load_from_sql()
        else:
            self.load_from_sparql()
            for record in self.lod:
                pid=record["pid"]
                lang=record["lang"]
                pid = pid.replace("http://www.wikidata.org/entity/", "")
                record["pid"]=pid
                record["id"]=f"{pid}-{lang}"
            self.store()
        self.init_props()
        self.loaded=True
            
    def init_props(self):
        """
        initialize my property structures
        """
        self.props = []
        self.props_by_id = {}
        self.props_by_lang = {}
        if not self.lod:
            raise Exception(f"Could not fetch wikidata properties for {self.langs}")
        for record in self.lod:
            prop = WikidataProperty(**record)
            self.props.append(prop)
        for lang in self.langs:
            self.props_by_lang[lang] = {}
            self.props_by_id[lang]={}
        for prop in self.props:
            self.props_by_lang[prop.lang][prop.plabel] = prop
            self.props_by_id[prop.lang][prop.pid] = prop

    def get_mappings_for_records(
        self, prop_mapping_records: Dict[str, dict]
    ) -> List["PropertyMapping"]:
        """
        convert given list of property mapping records to list of PropertyMappings
        Args:
            prop_mapping_records: records to convert

        Returns:
            property mappings
        """
        mappings = []
        for record in prop_mapping_records.values():
            mapping = PropertyMapping.from_record(self, record)
            mappings.append(mapping)
        return mappings

    def get_query_for_langs(self, langs: list = None) -> str:
        """
        Get the SPARQL query for the given list of langs.
        """
        query_prefix = Prefixes.getPrefixes(["wikibase", "rdfs", "schema"])
        query_body = ""
        if langs is None:
            langs = self.langs
        for lang in langs:
            if query_body:  # If not the first iteration, add UNION
                query_body += "\n  UNION"
            query_body += f"""
  {{ # wikidata properties with {lang} labels and descriptions
    ?property a wikibase:Property;
    rdfs:label ?propertyLabel;
    schema:description ?propertyDescription;
    wikibase:propertyType ?wbType.
    FILTER(LANG(?propertyLabel) = "{lang}") .
    FILTER(LANG(?propertyDescription) = "{lang}") .
    BIND("{lang}" AS ?lang)
  }}"""
        query = (
            query_prefix + "SELECT \n"
            "  (STR(?property) AS ?pid)\n"
            "  ?lang\n"
            "  (?propertyLabel AS ?plabel)\n"
            "  (?propertyDescription AS ?description)\n"
            "  (STR(?wbType) AS ?type_name)\n"
            "WHERE {" + query_body + "\n}\n"
        )
        return query

    @classmethod
    def get_instance(
        cls,
        endpoint_url: str = "https://qlever.cs.uni-freiburg.de/api/wikidata",
    ) -> "WikidataPropertyManager":
        """
        initialize the wikidata property manager

        Args:
            endpoint_url(str): the SPARQL endpoint to query if there is no cache available
            lang(str): the languages to query propery labels and descriptions for
        """
        if not hasattr(cls, "wpm"):
            cls.wpm = WikidataPropertyManager(endpoint_url)
        return cls.wpm

    @classmethod
    def get_cache_path(cls, lang: str = "en") -> str:
        home = str(Path.home())
        cache_dir = f"{home}/.wikidata"
        os.makedirs(cache_dir, exist_ok=True)
        cache_path = f"{cache_dir}/wikidata_properties.db"
        return cache_path

    def get_properties_by_labels(
        self, labels: List[str], lang: str = "en"
    ) -> Dict[str, WikidataProperty]:
        """
        Get properties by their labels for a specific language.

        Args:
            labels: List of property labels to search for.
            lang: the language to match with
        Returns:
            A dictionary of {label: WikidataProperty} for found properties.
        """
        matched_properties = {}
        # Check if language exists in cached properties
        # Iterate over requested labels and try to find them in the cached properties
        for label in labels:
            if label in self.props_by_lang[lang]:
                matched_properties[label] = self.props_by_lang[lang][label]
        return matched_properties

    def get_properties_by_ids(
        self, ids: List[str],lang:str="en"
    ) -> Dict[str, Optional[WikidataProperty]]:
        """
        Get properties by their IDs for a specific language.

        Args:
            ids: List of property IDs to search for.
            lang(str): the language

        Returns:
            A dictionary of {property ID: WikidataProperty or None} for found and not found properties.
        """
        matched_properties = {}
        for pid in ids:
            # first check requested language
            if pid in self.props_by_id[lang]:
                matched_properties[pid] = self.props_by_id[lang][pid]
            elif pid in self.props_by_lang["en"]:
                # fall back to english
                matched_properties[pid] = self.props_by_id["en"][pid]
        return matched_properties

    def get_property_by_id(self, property_id: str) -> WikidataProperty:
        """
        lookup a WikidataProperty for the given property_id

        Args:
            property_id(str): a property ID e.g. "P6375"
        """
        properties = self.get_properties_by_ids([property_id])
        prop_count = len(properties)
        if prop_count == 1:
            return list(properties.values())[0]
        elif prop_count == 0:
            return None
        else:
            property_labels = list(properties.keys())
            msg = f"unexpected get_property_by_id result for property id {property_id}. Expected 0 or 1 results bot got:{property_labels}"
            raise ValueError(msg)
        pass


@lod_storable
class PropertyMapping:
    """
    Represents a single column Wikidata property mapping.

    Attributes:
        column (Optional[str]): The column name in the data source; if None, the value is directly used.
        propertyName (str): The human-readable name of the property.
        propertyId (str): The Wikidata property ID (e.g., "P31").
        propertyType (str): The type of the property as a string; converted to an enum in post-init.
        qualifierOf (Optional[str]): Specifies if the property is a qualifier of another property.
        valueLookupType (Optional[Any]): The type (instance of/P31) of the property value for lookup if the value is not already a QID.
        value (Optional[Any]): The default value to set for the property.
        varname (Optional[str]): An optional variable name for internal use.
        property_type_enum (WdDatatype): The enum representation of the property type, initialized based on propertyType.

    The __post_init__ method ensures the propertyType is correctly interpreted and stored as both a string and an enum.
    """

    column: Union[str, None]  # if None, the value is used
    propertyName: str
    propertyId: str
    propertyType: str
    qualifierOf: str = None
    valueLookupType: Any = None  # type (instance of/P31) of the property value → used to lookup the qid if property value if value is not already a qid
    value: Any = None  # set this value for the property
    varname: str = None
    # property_type_enum: WdDatatype=field(init=False)

    def __post_init__(self):
        """
        Convert propertyType from string to WdDatatype enum if necessary
        """
        self.property_type_enum = None
        if isinstance(self.propertyType, str):
            try:
                self.property_type_enum = WdDatatype[self.propertyType]
            except KeyError:
                raise ValueError(f"Invalid property type: {self.propertyType}")
        else:
            self.property_type_enum = self.propertyType
            # Ensure propertyType is stored as the correct string representation of the enum for YAML compatibility
            self.propertyType = self.property_type_enum.name

    @classmethod
    def get_legacy_mapping(cls) -> dict:
        """
        Returns the Mapping from old prop map keys to the new once
        """
        return {
            "Column": "column",
            "PropertyName": "propertyName",
            "PropertyId": "propertyId",
            "Type": "propertyType",
            "Qualifier": "qualifierOf",
            "Lookup": "valueLookupType",
            "Value": "value",
            "PropVarname": "varname",
        }

    @classmethod
    def from_record(
        cls, wpm: WikidataPropertyManager, record: dict
    ) -> "PropertyMapping":
        """
        initialize PropertyMapping from the given record
        Args:
            wpm(WikidataPropertyManager): to be used for type lookup
            record(Dict): property mapping information

        Returns:
            PropertyMapping
        """
        legacy_lookup = cls.get_legacy_mapping()
        record = record.copy()
        for i in range(len(record)):
            key = list(record.keys())[i]
            if key in legacy_lookup:
                record[legacy_lookup[key]] = record[key]
        # handle missing property type
        property_type = record.get("propertyType", None)
        if property_type in [None, ""]:
            if record.get("valueLookupType", None) not in [None, ""]:
                property_type = WdDatatype.itemid
            elif record.get("value", None) not in [None, ""]:
                property_type = WdDatatype.itemid
        if property_type is not None and not isinstance(property_type, WdDatatype):
            if property_type in [wd.name for wd in WdDatatype]:
                property_type = WdDatatype[property_type]
            else:
                pid = record.get("propertyId")
                props = wpm.get_properties_by_ids([pid])
                if len(props) == 1:
                    prop = list(props.values())[0]
                    property_type = prop.ptype
        mapping = PropertyMapping(
            column=record.get("column", None),
            propertyName=record.get("propertyName", None),
            propertyId=record.get("propertyId", None),
            propertyType=property_type,
            qualifierOf=record.get("qualifierOf", None),
            valueLookupType=record.get("valueLookupType", None),
            value=record.get("value", None),
            varname=record.get("varname", None),
        )
        return mapping

    def to_record(self) -> dict:
        """
        convert property mapping to its dict representation
        """
        key_map = self.get_legacy_mapping()
        record = dict()
        for old_key, new_key in key_map.items():
            record[old_key] = getattr(self, new_key, None)
        return record

    def is_qualifier(self) -> bool:
        """
        Returns true if the property mapping describes a qualifier
        """
        is_qualifier = not (self.qualifierOf is None or self.qualifierOf == "")
        return is_qualifier

    @classmethod
    def getDefaultItemPropertyMapping(cls) -> "PropertyMapping":
        """
        get the defaultItemPropertyMapping
        """
        if not hasattr(cls, "defaultItemPropertyMapping"):
            item_prop_map = PropertyMapping(
                column="item",
                propertyName="item",
                propertyId="",
                propertyType=WdDatatype.item,
                varname="item",
            )
            cls.defaultItemPropertyMapping = item_prop_map
        return cls.defaultItemPropertyMapping

    def is_item_itself(self) -> bool:
        """
        Check if the property_type is an item

        Returns:
            bool: True if the property mapping links to the existing item
        """
        is_item_id = self.property_type_enum == WdDatatype.item
        return is_item_id

    @classmethod
    def get_qualifier_lookup(
        cls, properties: List["PropertyMapping"]
    ) -> Dict[str, List["PropertyMapping"]]:
        """
        Get a lookup for a property and all its qualifier

        Args:
            properties: property mappings to generate the lookup from

        Returns:
             dict as property qualifier lookup
        """
        res = dict()
        for pm in properties:
            if not isinstance(pm, PropertyMapping):
                continue
            if pm.qualifierOf is None or pm.qualifierOf == "":
                continue
            else:
                if pm.qualifierOf in res:
                    res[pm.qualifierOf].append(pm)
                else:
                    res[pm.qualifierOf] = [pm]
        return res

    @classmethod
    def get_item_mapping(
        cls, property_mappings: List["PropertyMapping"]
    ) -> "PropertyMapping":
        """
        get the property mapping that is used for the default "item" primary key
        if no property is defined use the default "item" mapping
        """
        for pm in property_mappings:
            if pm.is_item_itself():
                return pm
        pm = cls.getDefaultItemPropertyMapping()
        return pm


@lod_storable
@dataclass
class PropertyMappings:
    """
    A collection of Wikidata property mappings, with metadata.
    """

    name: str
    mappings: Dict[str, PropertyMapping] = field(default_factory=dict)
    description: Optional[str] = None
    url: Optional[str] = None



"""
Created on 2022-03-4

@author: wf
"""
import json
import unittest
from pprint import pprint
from urllib.error import HTTPError

from lodstorage.query import Endpoint, Query, QuerySyntaxHighlight
from lodstorage.sparql import SPARQL

from ez_wikidata.trulytabular import TrulyTabular
from ez_wikidata.wdproperty import WikidataPropertyManager
from ez_wikidata.wikidata import WikidataItem


class TestTrulyTabular(unittest.TestCase):
    """
    test Truly tabular analysis
    """

    def setUp(self):
        self.debug = False
        qleverEndpoint = Endpoint()
        qleverEndpoint.name = "qlever-wikidata"
        qleverEndpoint.method = "POST"
        qleverEndpoint.database = "qlever"
        # RWTH Aachen i5 endpoint
        qleverEndpoint.endpoint = "http://qlever-api.wikidata.dbis.rwth-aachen.de"

        self.endpointConfs = {qleverEndpoint, Endpoint.getDefault()}
        pass

    def handleServiceUnavailable(self, ex, endpointConf):
        """
        handle service unavailable

        Args:
            ex(Exception): the exception to handle
            endpointConf(Endpoint): the endpoint for which there is a problem
        """
        self.handleEndpointErrors(ex, endpointConf, "503", "Service Unavailable")

    def handleEndpointErrors(
        self, ex, endpointConf, status_code: str, status_text: str
    ):
        """
         handle Endpoint Errors

        Args:
            ex(Exception): the exception to handle
            endpointConf(Endpoint): the endpoint for which there is a problem
            status_code(str): the status code to filter
            status_text(str): the description of the status code
        """
        if status_code in str(ex):
            print(
                f"{endpointConf.name} at {endpointConf.endpoint} returns {status_code} ({status_text})",
                flush=True,
            )
        else:
            raise (ex)

    def tearDown(self):
        pass

    def testGetFirst(self):
        """
        test the get First helper function
        """
        tt = TrulyTabular("Q2020153")
        testcases = [
            {"qlod": [{"name": "firstname"}], "expected": "firstname"},
            {"qlod": [], "expected": None},
            {
                "qlod": [{"name": "firstname"}, {"name": "second name"}],
                "expected": None,
            },
        ]
        for testcase in testcases:
            qLod = testcase["qlod"]
            expected = testcase["expected"]
            try:
                value = tt.sparql.getFirst(qLod, "name")
                self.assertEqual(expected, value)
            except Exception as ex:
                if self.debug:
                    print(str(ex))
                self.assertIsNone(expected)

    def documentQuery(self, tt, query, show=True, formats=["mediawiki"]):
        """
        document the given query for the given TrueTabular instance
        """
        qlod = tt.sparql.queryAsListOfDicts(query.query)
        for tablefmt in formats:
            tryItUrl = "https://query.wikidata.org/"
            doc = query.documentQueryResult(
                qlod, tablefmt=tablefmt, tryItUrl=tryItUrl, floatfmt=".0f"
            )
            docstr = doc.asText()
            if show:
                print(docstr)

    def testGetPropertiesByLabel(self):
        """
        try getting properties by label
        """
        debug = self.debug
        # debug=True
        propertyLabels = ["title", "country", "location"]
        for endpointConf in self.endpointConfs:
            try:
                tt = TrulyTabular(
                    "Q2020153", propertyLabels=propertyLabels, endpointConf=endpointConf
                )
                if debug:
                    print(tt.properties)
                for prop in propertyLabels:
                    self.assertTrue(prop in tt.properties)
            except (Exception, HTTPError) as ex:
                self.handleServiceUnavailable(ex, endpointConf)
                pass

    def testGetPropertiesById(self):
        """
        try getting properties by label
        """
        debug = self.debug
        # debug=True
        propertyIds = ["P1800"]
        expected = ["Wikimedia database name"]
        for endpointConf in self.endpointConfs:
            try:
                wpm = WikidataPropertyManager.get_instance(
                    endpoint_url=endpointConf.endpoint
                )
                prop_dict = wpm.get_properties_by_ids(
                    propertyIds,
                )
                for i, prop in enumerate(prop_dict.values()):
                    if debug:
                        print(f"{endpointConf.name} {i}:{prop}")
                    self.assertEqual(prop.plabel, expected[i])
            except (Exception, HTTPError) as ex:
                self.handleServiceUnavailable(ex, endpointConf)
                pass

    def testGetItemsByLabel(self):
        """
        try getting items by label
        """
        debug = self.debug
        # debug = True
        qLabels = [
            "academic conference",
            "scientific conference series",
            "whisky distillery",
            "human",
        ]
        for endpointConf in self.endpointConfs:
            try:
                sparql = SPARQL(endpointConf.endpoint, method=endpointConf.method)
                items = {}
                for qLabel in qLabels:
                    items4Label = WikidataItem.getItemsByLabel(
                        sparql, qLabel, debug=debug
                    )
                    count = len(items4Label)
                    if debug:
                        print(f"found {count} items for label {qLabel}")
                    self.assertTrue(count > 0)
                    for i, item in enumerate(items4Label):
                        if debug:
                            print(f"{endpointConf.name} {i+1}:{item}")
                    items[qLabel] = items4Label[0]
                for qLabel in qLabels:
                    self.assertTrue(qLabel in items)
            except (Exception, HTTPError) as ex:
                self.handleServiceUnavailable(ex, endpointConf)
                pass

    def testTrulyTabularTables(self):
        """
        test Truly Tabular for different tabular queries
        """
        debug = self.debug
        # debug=True
        show = False
        showStats = ["mediawiki", "github", "latex"]
        tables = [
            {
                "name": "computer scientist",
                "title": "humans with the occupation computer scientist",
                "qid": "Q5",  # human
                "where": "?item wdt:P106 wd:Q82594.",  # computer scientist only
                "propertyLabels": [
                    "sex or gender",
                    "date of birth",
                    "place of birth",
                    "field of work",
                    "occupation",
                    "ORCID iD",
                    "GND ID",
                    "DBLP author ID",
                    "Google Scholar author ID",
                    "VIAF ID",
                ],
                "expected": 10,
            },
            {
                "name": "academic conferences",
                "title": "academic conferences",
                "qid": "Q2020153",  # academic conference
                "propertyLabels": [
                    "title",
                    "country",
                    "location",
                    "short name",
                    "start time",
                    "end time",
                    "part of the series",
                    "official website",
                    "described at URL",
                    "WikiCFP event ID",
                    "GND ID",
                    "VIAF ID",
                    "main subject",
                    "language used",
                    "is proceedings from",
                ],
                "expected": 7500,
            },
            {
                "name": "scientific conferences series",
                "title": "scientific conference series",
                "qid": "Q47258130",  # scientific conference series
                "propertyLabels": [
                    "title",
                    "short name",
                    "inception",
                    "official website",
                    "DBLP venue ID",
                    "GND ID",
                    "Microsoft Academic ID",
                    "Freebase ID",
                    "WikiCFP conference series ID",
                    "Publons journals/conferences ID",
                    "ACM conference ID",
                ],
                "expected": 4200,
            },
            {
                "name": "whisky distilleries",
                "title": "whisky distilleries",
                "qid": "Q10373548",  # whisky distillery
                "propertyLabels": [
                    "inception",
                    "official website",
                    "owned by",
                    "country",
                    "headquarters location",
                    "Whiskybase distillery ID",
                ],
                "expected": 200,
            },
        ]
        errors = 0
        for table in tables[3:]:
            # academic conference
            where = None
            if "where" in table:
                where = table["where"]
            tt = TrulyTabular(
                table["qid"], table["propertyLabels"], where=where, debug=debug
            )
            if "is proceedings from" in tt.properties:
                tt.properties["is proceedings from"].reverse = True
            count, query = tt.count()
            if debug:
                print(count)
            self.assertTrue(count > table["expected"])
            stats = tt.getPropertyStatistics()
            # sort descending by total percentage
            stats = sorted(stats, key=lambda row: row["total%"], reverse=True)
            for tablefmt in showStats:
                query = Query(name=table["name"], title=table["title"], query="")
                doc = query.documentQueryResult(
                    stats, tablefmt=tablefmt, withSourceCode=False
                )
                if debug:
                    print(doc)
            if show:
                for wdProperty in tt.properties.values():
                    for asFrequency in [True, False]:
                        query = tt.noneTabularQuery(wdProperty, asFrequency=asFrequency)
                        try:
                            self.documentQuery(tt, query)
                        except Exception as ex:
                            print(f"query for {wdProperty} failed\n{str(ex)}")
                            errors += 1
                self.assertEqual(0, errors)

    def testMostFrequentProperties(self):
        """
        test getting the most frequent properties for some Wikidata Item types
        """
        # show=True
        show = False
        debug = self.debug
        # debug=True
        for endpointConf in self.endpointConfs:
            for qid in ["Q6256"]:
                try:
                    tt = TrulyTabular(qid, debug=debug, endpointConf=endpointConf)
                    for minCount in [0, 100]:
                        query = tt.mostFrequentPropertiesQuery(minCount=minCount)
                        if debug:
                            print(query.query)
                        self.documentQuery(tt, query, formats=["github"], show=show)
                except (Exception, HTTPError) as ex:
                    self.handleServiceUnavailable(ex, endpointConf)
                    pass

    def testSyntaxHighlighting(self):
        """
        https://github.com/WolfgangFahl/pyLoDStorage/issues/81
        """
        debug = self.debug
        # debug=True
        qid = "Q6256"  # country
        tt = TrulyTabular(qid, debug=debug)
        query = tt.mostFrequentPropertiesQuery()
        sh = QuerySyntaxHighlight(query, "html")
        html = sh.highlight()
        if debug:
            print(html)
        self.assertTrue('<span class="k">SELECT</span>' in html)
        pass

    def testCount(self):
        """
        test the count function of truly tabular
        """
        debug = self.debug
        # debug = True
        qid = "Q55488"  # railway stations
        for endpointConf in self.endpointConfs:
            try:
                tt = TrulyTabular(qid, endpointConf=endpointConf, debug=debug)
                count, query = tt.count()
                if debug:
                    print(query)
                    print(f"count of railway stations is {count}")
                self.assertTrue(qid in query)
                self.assertTrue(count >= 106195)
                self.assertTrue(tt.error is None)
            except (Exception, HTTPError) as ex:
                self.handleServiceUnavailable(ex, endpointConf)

    def testGenWdPropertyStatistic(self):
        """
        test generating a Wikidata property statistics row
        """
        qid = "Q44613"  # monastery
        debug = self.debug
        # debug = True
        for endpointConf in self.endpointConfs:
            wpm = WikidataPropertyManager.get_instance(
                endpoint_url=endpointConf.endpoint
            )
            try:
                tt = TrulyTabular(qid, debug=debug)
                # P571: inception, P6375: street address
                for pid in ["P571", "P6375"]:
                    wdProperty = wpm.get_property_by_id(pid)
                    self.assertIsNotNone(wdProperty)
                    #
                    itemCount, _itemCountQuery = tt.count()
                    statsRow = tt.genWdPropertyStatistic(wdProperty, itemCount)
                    if debug:
                        print(json.dumps(statsRow, indent=2))
            except (Exception, HTTPError) as ex:
                self.handleEndpointErrors(ex, endpointConf, "405", "Method not allowed")

    def testGenerateSparqlQuery(self):
        """
        test Generating a SPARQL query
        """
        configs = [
            {
                "naive": True,
                "qid": "Q2020153",  # academic conference
                "search_predicate": "wdt:P31",
                "propertyIdMap": {
                    "P1813": ["label"],
                    "P17": ["label"],
                    "P1476": ["label"],
                },
                "expected": [],
            },
            {
                "naive": False,
                "qid": "Q2020153",  # academic conference
                "search_predicate": "wdt:P31",
                "propertyIdMap": {
                    "P1813": ["sample"],
                    "P17": ["sample"],
                    "P1476": ["sample"],
                },
                "expected": ["GROUP BY", "SAMPLE"],
            },
            {
                "naive": False,
                "qid": "Q2020153",  # academic conference
                "search_predicate": "wdt:P31",
                "propertyIdMap": {
                    "P1813": ["count", "list"],
                    "P17": ["sample", "ignore"],
                    "P1476": ["count", "list"],
                },
                "expected": [
                    "COUNT (DISTINCT",
                    "GROUP BY",
                    "GROUP_CONCAT (DISTINCT",
                    "HAVING",
                ],
            },
            {
                "naive": False,
                "qid": "Q1667921",  # novel series
                "search_predicate": "wdt:P31",
                "propertyIdMap": {
                    "P50": ["sample", "ignore"],  # author
                    "P136": ["sample", "ignore"],  # genre
                    "P1476": ["sample", "ignore"],  # title
                },
                "expected": ["GROUP BY", "HAVING", "COUNT", "<=1"],
            },
            {
                "naive": False,
                "qid": "Q1667921",  # novel series
                "search_predicate": "wdt:P31",
                "propertyIdMap": {
                    "P50": ["sample", "ignore", "label"],  # author
                    "P136": ["sample", "ignore", "label"],  # genre
                    "P1476": ["sample", "ignore"],  # title
                },
                "expected": ["GROUP BY", "HAVING", "COUNT", "<=1"],
            },
            {
                "naive": False,
                "search_predicate": "wdt:P279*/wdt:P31*",
                "qid": "Q8063",  # rock
                "propertyIdMap": {
                    "P18": ["sample"],  # image
                },
                "expected": ["P279"],
            },
        ]
        debug = self.debug
        # debug = True
        # loop over different test configurations
        for i, config in enumerate(configs):
            # get the test configuration
            qid = config["qid"]
            naive = config["naive"]
            propertyIdMap = config["propertyIdMap"]
            search_predicate = config["search_predicate"]
            expectedList = config["expected"]

            # create a truly tabular analysis
            tt = TrulyTabular(
                qid,
                propertyIds=list(propertyIdMap.keys()),
                search_predicate=search_predicate,
            )
            varname = tt.item.itemVarname
            # generate a SPARQL Query
            sparqlQuery = tt.generateSparqlQuery(genMap=propertyIdMap, naive=naive)
            if debug:
                print(f"config {i}:")
                pprint(config)
                print(f"{sparqlQuery}")
            # all queries should have basic graph patterns for the subclass
            self.assertTrue(f"?{varname} {search_predicate} wd:{qid}." in sparqlQuery)
            # and for the properties
            for pid in propertyIdMap.keys():
                self.assertTrue(f"?{varname} wdt:{pid}" in sparqlQuery)
            for expected in expectedList:
                self.assertTrue(
                    expected in sparqlQuery, f"config {i}:{expected} missing"
                )



"""
Created on 2024-03-01

@author: wf
"""
from ez_wikidata.wbquery import WikibaseQuery
from tests.basetest import BaseTest


class TestWikibaseQuery(BaseTest):
    """
    test the WikibaseQuery module
    """

    def setUp(self, debug=False, profile=True):
        super().setUp(debug, profile)

    def test_init(self):
        """Test initialization of WikibaseQuery."""
        query = WikibaseQuery(entity="Q5", debug=True)
        self.assertTrue(query.debug)
        self.assertEqual(query.entity, "Q5")
        self.assertEqual(len(query.propertiesByName), 0)

    def test_addPropertyFromDescriptionRow(self):
        """
        Test adding a property from a description
        row with entity beer and property Country."""
        query = WikibaseQuery(entity="Q44")  # Assuming Q44 is the entity ID for "beer"
        # This example assumes the structure of your row data. Adjust as necessary.
        row = {
            "PropertyName": "country",  # Human-readable name for the property
            "PropertyId": "P17",  # Property ID for "country"
            "Column": "CountryCol",  # The column name where this property's data will be stored or displayed
            "Type": "item",  # Assuming the country is represented as an item in Wikidata
        }
        query.addPropertyFromDescriptionRow(row)

        self.assertIn("country", query.propertiesByName)
        self.assertIn("P17", query.propertiesById)
        self.assertIn("CountryCol", query.propertiesByColumn)
        self.assertIn("country", query.propertiesByVarname)



import json
import unittest
import uuid
from datetime import datetime

from lodstorage.lod import LOD
from wikibaseintegrator.datatypes import (
    URL,
    ExternalID,
    Item,
    MonolingualText,
    String,
    Time,
)

from ez_wikidata.wikidata import (
    PropertyMapping,
    PropertyMappings,
    UrlReference,
    WdDatatype,
    Wikidata,
    WikidataItem,
    WikidataResult,
)
from tests.basetest import BaseTest


class TestWikidata(BaseTest):
    """
    test the Wikidata access
    """

    def setUp(self, debug=False, profile=True):
        super().setUp(debug, profile)
        self.wd = Wikidata("https://www.wikidata.org", debug=debug)
        self._test_wd = None

    @property
    def test_wikidata(self) -> Wikidata:
        if self._test_wd is None:
            wd = Wikidata(baseurl="https://test.wikidata.org", debug=self.debug)
            wd.loginWithCredentials()
            self._test_wd = wd
        return self._test_wd

    def testItemLookup(self):
        """
        lookup items
        """
        debug = self.debug
        wd = Wikidata("https://www.wikidata.org", debug=self.debug)
        test_params = [
            ("USA", "Q3624078", "Q30"),
            ("ECIR", "Q47258130", "Q5412436"),
            ("Berlin", "Q515", "Q64"),
        ]
        for test_param in test_params:
            with self.subTest(test_param=test_param):
                name, item_type, expected_qid = test_param
                actual_qid = wd.getItemByName(name, item_type)
                if debug:
                    print(actual_qid)
                self.assertEqual(expected_qid, actual_qid)

    def test_convert_to_claim(self):
        """
        tests convert_to_claim
        """
        acronym = PropertyMapping(
            column="acronym",
            propertyId="P1813",
            propertyName="short name",
            propertyType=WdDatatype.text,
        )
        homepage = PropertyMapping(
            column="homepage",
            propertyId="P856",
            propertyName="official website",
            propertyType=WdDatatype.url,
        )
        inception = PropertyMapping(
            column="inception",
            propertyId="P571",
            propertyName="inception",
            propertyType=WdDatatype.year,
        )
        start_time = PropertyMapping(
            column="startTime",
            propertyId="P580",
            propertyName="start time",
            propertyType=WdDatatype.date,
        )
        issn = PropertyMapping(
            column="issn",
            propertyId="P236",
            propertyName="ISSN",
            propertyType=WdDatatype.extid,
        )
        country = PropertyMapping(
            column="country",
            propertyId="P17",
            propertyName="country",
            propertyType=WdDatatype.itemid,
        )
        volume = PropertyMapping(
            column="volume",
            propertyId="P478",
            propertyName="volume",
            propertyType=WdDatatype.string,
        )
        test_params = [  # (value, expected value, expected  type, property mapping)
            ("AAAI", {"text": "AAAI", "language": "en"}, MonolingualText, acronym),
            ("http://ceur-ws.org/", "http://ceur-ws.org/", URL, homepage),
            (
                "1995",
                {
                    "time": "+1995-01-01T00:00:00Z",
                    "before": 0,
                    "after": 0,
                    "precision": 9,
                    "timezone": 0,
                    "calendarmodel": "http://www.wikidata.org/entity/Q1985727",
                },
                Time,
                inception,
            ),
            (
                "1995-05-04",
                {
                    "time": "+1995-05-04T00:00:00Z",
                    "before": 0,
                    "after": 0,
                    "precision": 11,
                    "timezone": 0,
                    "calendarmodel": "http://www.wikidata.org/entity/Q1985727",
                },
                Time,
                start_time,
            ),
            ("1613-0073", "1613-0073", ExternalID, issn),
            (
                "Q30",
                {"entity-type": "item", "numeric-id": 30, "id": "Q30"},
                Item,
                country,
            ),
            ("1", "1", String, volume),
            ("A", "A", String, volume),
            (1, "1", String, volume),
        ]
        for params in test_params:
            with self.subTest("test statement generation", params=params):
                value, expected_value, expected_type, pm = params
                statement = self.wd.convert_to_claim(value, pm)
                self.assertTrue(isinstance(statement, expected_type))
                self.assertEqual(expected_value, statement.mainsnak.datavalue["value"])

    def test_is_wikidata_item_id(self):
        """
        tests is_wikidata_item_id
        """
        test_params = [("human", False), ("Q5", True), ("P17", False)]
        for param in test_params:
            with self.subTest(param=param):
                value, expected = param
                actual = self.wd.is_wikidata_item_id(value)
                self.assertEqual(expected, actual)

    def test_is_wikidata_property_id(self):
        """
        tests is_wikidata_item_id
        """
        test_params = [("country", False), ("P17", True), ("Q5", False)]
        for param in test_params:
            with self.subTest(param=param):
                value, expected = param
                actual = self.wd.is_wikidata_property_id(value)
                self.assertEqual(expected, actual)

    def test_sanitize_label(self):
        """
        tests sanitize_label
        """
        test_params = [
            ("Hello World", 8, None, "Hello..."),
            ("A" * 255, None, None, "A" * 247 + "..."),
            ("P17", 2, "!", "P!"),
        ]
        for param in test_params:
            with self.subTest(param=param):
                label, limit, postfix, expected = param
                actual = self.wd.sanitize_label(label, limit, postfix)
                self.assertEqual(expected, actual)

    def test_get_record(self):
        """
        tests get_record
        ToDo: add more test cases
        """
        expected = {
            "P31": "Q1143604",
            "P1813": "JIST-WP 2017",
            "P577": datetime(2017, 11, 20).date(),
            "P973": "http://ceur-ws.org/Vol-2000/",
            "P0": None,  # invalid prop id to test invalid key handling
        }
        prop_ids = list(expected.keys())
        actual = self.wd.get_record(
            "Q113543868", prop_ids, include_label=False, include_description=False
        )
        self.assertDictEqual(expected, actual)

    def test_get_record_label(self):
        """
        test get_record with property mapping to extract the label of item links
        """
        pm = PropertyMapping(
            column="country",
            propertyName="country",
            propertyType=WdDatatype.itemid,
            propertyId="P17",
            valueLookupType="Q3624078",
        )
        item_id = "Q64"  # Berlin
        expected = {"country": "Germany"}
        actual = self.wd.get_record(
            item_id,
            property_mappings=[pm],
            include_label=False,
            include_description=False,
            label_for_qids=True,
        )
        self.assertDictEqual(expected, actual)

    def check_add_result(self, result: WikidataResult):
        """
        check the result of an add operation
        """
        if result.debug:
            if len(result.errors) > 0:
                print("Errors")
                for index, error in enumerate(result.errors.values()):
                    print(f"{index+1}:{str(error)}")
            else:
                print(f"created {result.qid}")
                print(result.pretty_item_json)
        self.assertEqual(0, len(result.errors))

    @unittest.skipIf(
        BaseTest.inPublicCI(),
        "Tests querying wikidata which is often blocked on public CI",
    )
    def test_get_item_label(self):
        """
        tests get_item_label
        """
        test_params = [
            ("Q64", "en", "Berlin"),
            ("Q183", "en", "Germany"),
            ("Q183", "de", "Deutschland"),
            ("Q64", "xxx", None),
            (None, "en", None),
            ("Q183", None, "Germany"),
        ]
        for test_param in test_params:
            with self.subTest(test_param=test_param):
                item_id, lang, expected_label = test_param
                actual_label = self.wd.get_item_label(item_id, lang)
                self.assertEqual(expected_label, actual_label)

    @unittest.skipIf(
        BaseTest.inPublicCI(),
        "Tests creating and modifying items. To run in CI setup credentials",
    )
    def test_add_record(self):
        """
        tests add_record by creating an item in test.wikidata.org adding property values.
        Also tests adding property values to existing items.
        """
        debug = self.debug
        debug = True
        wd = Wikidata(baseurl="https://test.wikidata.org", debug=debug)
        wd.loginWithCredentials()
        property_mappings = [
            WikidataSandboxProperties.TEXT,
            WikidataSandboxProperties.ITEM,
            WikidataSandboxProperties.DATE,
            WikidataSandboxProperties.YEAR,
            WikidataSandboxProperties.URL,
            WikidataSandboxProperties.EXT_ID,
            PropertyMapping(
                column="dateQualifier",
                propertyId="P96927",
                propertyName="year sandbox property",
                propertyType=WdDatatype.year,
                qualifierOf="date",
            ),
        ]
        record = {
            "label": str(uuid.uuid4()),
            "description": "test item added to test correctness of api",
            "text": "test",
            "date": datetime.now().date(),
            "dateQualifier": datetime.now().year,
            "item_id": "Q377",
            "url": "https://example.org",
            "year": 2000,
            "identifier": str(uuid.uuid4()),
        }

        # test creating an item
        result = wd.add_record(
            record=record, property_mappings=property_mappings, write=True
        )
        self.check_add_result(result)

        actual = wd.get_record(result.qid, property_mappings=property_mappings)
        self.assertDictEqual(record, actual)

        # test modifying an item (not overwriting existing value)
        record["year"] = [record["year"], 2022]
        result = wd.add_record(
            {"year": 2022},
            property_mappings=property_mappings,
            item_id=result.qid,
            write=True,
        )
        self.check_add_result(result)
        actual = wd.get_record(result.qid, property_mappings=property_mappings)
        self.assertDictEqual(record, actual)

    @unittest.skipIf(
        BaseTest.inPublicCI(),
        "Tests creating and modifying items. To run in CI setup credentials",
    )
    def test_addDict(self):
        """
        test addDict
        """
        debug = self.debug
        debug = True
        wd = Wikidata(baseurl="https://test.wikidata.org", debug=debug)
        wd.loginWithCredentials()
        legacy_mappings = [
            {
                "Entity": "proceedings",
                "Column": None,
                "PropertyName": "instanceof",
                "PropertyId": "P95201",
                "Value": "Q377",
                "Type": "itemid",
                "Qualifier": None,
                "Lookup": None,
            },
            {
                "Entity": "proceedings",
                "Column": "short name",
                "PropertyName": "short name",
                "PropertyId": "P95227",
                "Type": "text",
                "Qualifier": None,
                "Lookup": "",
            },
            {
                "Entity": "proceedings",
                "Column": "pubDate",
                "PropertyName": "publication date",
                "PropertyId": "P95226",
                "Type": "date",
                "Qualifier": None,
                "Lookup": "",
            },
            {
                "Entity": "proceedings",
                "Column": "url",
                "PropertyName": "described at URL",
                "PropertyId": "P95231",
                "Type": "url",
                "Qualifier": None,
                "Lookup": "",
            },
            {
                "Entity": "proceedings",
                "Column": "language of work or name",
                "PropertyName": "language of work or name",
                "PropertyId": "P82",
                "Type": "itemid",
                "Qualifier": "url",
                "Lookup": "",
            },
            {
                "Entity": "proceedings",
                "Column": "urn",
                "PropertyName": "URN-NBN",
                "PropertyId": "P95232",
                "Type": "extid",
                "Qualifier": None,
                "Lookup": "",
            },
        ]
        mappings = [
            PropertyMapping(
                "instanceof",
                "instanceof",
                "P95201",
                propertyType=WdDatatype.itemid,
                value="Q1143604",
            ),
            PropertyMapping(
                "short name", "short name", "P95227", propertyType=WdDatatype.text
            ),
            PropertyMapping(
                "pubDate", "publication date", "P95226", propertyType=WdDatatype.date
            ),
            PropertyMapping(
                "url", "described at URL", "P95231", propertyType=WdDatatype.url
            ),
            PropertyMapping(
                "language of work or name",
                "language of work or name",
                "P82",
                propertyType=WdDatatype.itemid,
                qualifierOf="url",
            ),
            PropertyMapping("urn", "URN-NBN", "P95232", propertyType=WdDatatype.extid),
        ]
        record = {
            "label": str(uuid.uuid4()),
            "short name": "test",
            "pubDate": datetime.now().date(),
            "url": "http://example.org",
            "language of work or name": "Q377",
            "urn": str(uuid.uuid4()),
        }
        mapDict, _ = LOD.getLookup(legacy_mappings, "PropertyId")
        result = wd.addDict(record, mapDict=mapDict, write=True)
        self.check_add_result(result)
        actual = wd.get_record(result.qid, mappings)
        record["instanceof"] = "Q377"
        self.assertEqual(record, actual)

    @unittest.skipIf(
        BaseTest.inPublicCI(),
        "Tests creating and modifying items. To run in CI setup credentials",
    )
    def test_value_lookup(self):
        """
        tests the lookup of wikidata ids from label value
        Note: Currently the lookup is always against wikidata. Changing this requires to adapt this test accordingly.
        """
        lookup_type, label, expected_qid = (
            "Q3336843",  # constituent country of the United Kingdom - Wikidata
            "Scotland",  # label to lookup
            "Q22",  # expected lookup result
        )  # type qid, label, qid
        mappings = [
            PropertyMapping(
                column="country_id",
                propertyName="wikibase-item sandbox property",
                propertyId="P95201",
                propertyType=WdDatatype.itemid,
                valueLookupType=lookup_type,
            )
        ]
        # let't try adding a dict
        entity_label = f"str(uuid.uuid4())"
        record = {"label": entity_label, "country_id": label}
        expected_record = {"label": entity_label, "country_id": expected_qid}

        debug = self.debug
        #debug = True
        wd = Wikidata(baseurl="https://test.wikidata.org", debug=debug)
        wd.loginWithCredentials()
        result = wd.add_record(record, mappings, write=True)
        self.check_add_result(result)
        actual = wd.get_record(result.qid, mappings)
        self.assertDictEqual(expected_record, actual)

    @unittest.skipIf(BaseTest.inPublicCI(), "Tests creating statement with two values")
    def test_claim_with_two_values(self):
        """
        Tests adding a statement with two values by handing over a list
        """
        property_mappings = [
            WikidataSandboxProperties.TEXT,
            WikidataSandboxProperties.ITEM,
            WikidataSandboxProperties.DATE,
            WikidataSandboxProperties.URL,
            WikidataSandboxProperties.EXT_ID,
        ]
        record = {
            "label": str(uuid.uuid4()),
            "description": "test item added to test correctness of api",
            "text": ["Hello", "World"],
            "date": [
                datetime.now().date(),
                datetime.fromisoformat("2000-01-01").date(),
            ],
            "url": ["http://example.org", "http://example.org/test"],
            "item_id": ["Q377", "Q344"],
            "identifier": [str(uuid.uuid4()), str(uuid.uuid4())],
        }
        result = self.test_wikidata.add_record(
            record, property_mappings=property_mappings, write=True
        )
        self.check_add_result(result)
        actual_record = self.test_wikidata.get_record(
            result.qid, property_mappings=property_mappings
        )
        self.assertDictEqual(record, actual_record)

    def test_get_datatype_of_property(self):
        """
        tests get_datatype_of_property
        """
        test_params = [
            ("P31", "WikibaseItem"),
            ("31", "WikibaseItem"),
            (31, "WikibaseItem"),
            ("P580", "Time"),
            ("P1813", "Monolingualtext"),
            ("P856", "Url"),
            ("P8978", "ExternalId"),
        ]
        for param in test_params:
            with self.subTest(param=param):
                pid, expected = param
                actual = self.wd.get_datatype_of_property(pid)
                self.assertEqual(expected, actual)

    def test_get_wddatatype_of_property(self):
        """
        tests get_datatype_of_property
        """
        test_params = [
            ("P31", WdDatatype.itemid),
            ("31", WdDatatype.itemid),
            (31, WdDatatype.itemid),
            ("P580", WdDatatype.date),
            ("P1813", WdDatatype.text),
            ("P856", WdDatatype.url),
            ("P8978", WdDatatype.extid),
            ("P478", WdDatatype.string),
        ]
        for param in test_params:
            with self.subTest(param=param):
                pid, expected = param
                actual = self.wd.get_wddatatype_of_property(pid)
                self.assertEqual(expected, actual)

    def test_UrlReference(self):
        """
        tests UrlReference
        """
        test_params = [("http://example.org", "2022-01-01")]
        for param in test_params:
            with self.subTest(param=param):
                url, date = param
                reference = UrlReference(url, date)
                json = reference.get_json()
                json_str = str(json)
                self.assertIn("P854", json_str)
                self.assertIn("P813", json_str)
                self.assertIn(url, json_str)
                self.assertIn(date, json_str)

    @unittest.skipIf(BaseTest.inPublicCI(), "querying wikidata endpoint")
    def test_normalize_records(self):
        """
        tests normalize_records
        """
        prop_maps = [
            PropertyMapping(
                column="country",
                propertyType=WdDatatype.itemid,
                propertyName="country",
                propertyId="P17",
                valueLookupType="Q3624078",
            ),
            PropertyMapping(
                column="city",
                propertyType=WdDatatype.itemid,
                propertyName="location",
                propertyId="P276",
            ),
            PropertyMapping(
                column="location",
                propertyType=WdDatatype.itemid,
                propertyName="location",
                propertyId="P276",
            ),
            PropertyMapping(
                column="loc",
                propertyType=WdDatatype.itemid,
                propertyName="location",
                propertyId="P276",
            ),
            PropertyMapping(
                column=None,
                propertyType=WdDatatype.itemid,
                propertyName="location",
                propertyId="P276",
                value="Q5",
            ),
        ]
        record = {"country": "Germany", "city": "Q64", "location": None, "loc": ""}
        exp_normalized_record = {
            "country": WikidataItem("Q183", "Germany"),
            "city": WikidataItem("Q64", "Berlin"),
            "location": None,
            "loc": None,
        }
        actual_normalized_record = self.wd.normalize_records(record, prop_maps)
        self.assertDictEqual(exp_normalized_record, actual_normalized_record)

    @unittest.skipIf(BaseTest.inPublicCI(), "querying wikidata endpoint")
    def test_get_wikidata_item(self):
        """
        tests get_wikidata_item
        """
        test_params = [
            ("Germany", "Q3624078", WikidataItem("Q183", "Germany")),
            ("Q183", None, WikidataItem("Q183", "Germany")),
        ]
        for test_param in test_params:
            with self.subTest(test_param=test_param):
                qid_or_label, item_type_qid, expected = test_param
                actual = self.wd.get_wikidata_item(qid_or_label, item_type_qid)
                self.assertEqual(expected, actual)

    def test_convert_to_yaml(self):
        """
        test converting the PropertyMappings to YAML
        """
        sb_dict = {
            "text_field": WikidataSandboxProperties.TEXT,
            "date_field": WikidataSandboxProperties.DATE,
            "item_field": WikidataSandboxProperties.ITEM,
            "url_field": WikidataSandboxProperties.URL,
            "year_field": WikidataSandboxProperties.YEAR,
            "extid_field": WikidataSandboxProperties.EXT_ID,
        }
        pm = PropertyMappings(
            name="sandbox_props",
            description="TSandbox Properties for Testing",
            url="https://test.wikidata.org",
            mappings=sb_dict,
        )

        yaml_str = pm.to_yaml()
        if self.debug:
            print(yaml_str)


class WikidataSandboxProperties:
    """
    Wikidata sandbox items to be used to add and modify items in test.wikidata.org
    """

    TEXT = PropertyMapping(
        column="text",
        propertyId="P95227",
        propertyName="monolingualtext sandbox property",
        propertyType=WdDatatype.text,
    )
    DATE = PropertyMapping(
        column="date",
        propertyId="P95226",
        propertyName="time sandbox property",
        propertyType=WdDatatype.date,
    )
    ITEM = PropertyMapping(
        column="item_id",
        propertyId="P95201",
        propertyName="wikibase-item sandbox property",
        propertyType=WdDatatype.itemid,
    )
    URL = PropertyMapping(
        column="url",
        propertyId="P95231",
        propertyName="url sandbox property",
        propertyType=WdDatatype.url,
    )
    YEAR = PropertyMapping(
        column="year",
        propertyId="P96927",
        propertyName="year sandbox property",
        propertyType=WdDatatype.year,
    )
    EXT_ID = PropertyMapping(
        column="identifier",
        propertyId="P95232",
        propertyName="external-id sandbox property",
        propertyType=WdDatatype.extid,
    )



"""
Created on 2024-03-02

@author: wf
"""


class Prefixes:
    """
    handle standard Prefixes
    """

    @classmethod
    def getPrefixes(
        cls, prefixes=["rdf", "rdfs", "schema", "wd", "wdt", "wikibase", "xsd"]
    ):
        prefixMap = {
            "bd": "<http://www.bigdata.com/rdf#>",
            "cc": "<http://creativecommons.org/ns#>",
            "dct": "<http://purl.org/dc/terms/>",
            "geo": "<http://www.opengis.net/ont/geosparql#>",
            "ontolex": "<http://www.w3.org/ns/lemon/ontolex#>",
            "owl": "<http://www.w3.org/2002/07/owl#>",
            "p": "<http://www.wikidata.org/prop/>",
            "pq": "<http://www.wikidata.org/prop/qualifier/>",
            "pqn": "<http://www.wikidata.org/prop/qualifier/value-normalized/>",
            "pqv": "<http://www.wikidata.org/prop/qualifier/value/>",
            "pr": "<http://www.wikidata.org/prop/reference/>",
            "prn": "<http://www.wikidata.org/prop/reference/value-normalized/>",
            "prov": "<http://www.w3.org/ns/prov#>",
            "prv": "<http://www.wikidata.org/prop/reference/value/>",
            "ps": "<http://www.wikidata.org/prop/statement/>",
            "psn": "<http://www.wikidata.org/prop/statement/value-normalized/>",
            "psv": "<http://www.wikidata.org/prop/statement/value/>",
            "rdf": "<http://www.w3.org/1999/02/22-rdf-syntax-ns#>",
            "rdfs": "<http://www.w3.org/2000/01/rdf-schema#>",
            "schema": "<http://schema.org/>",
            "skos": "<http://www.w3.org/2004/02/skos/core#>",
            "wd": "<http://www.wikidata.org/entity/>",
            "wdata": "<http://www.wikidata.org/wiki/Special:EntityData/>",
            "wdno": "<http://www.wikidata.org/prop/novalue/>",
            "wdref": "<http://www.wikidata.org/reference/>",
            "wds": "<http://www.wikidata.org/entity/statement/>",
            "wdt": "<http://www.wikidata.org/prop/direct/>",
            "wdtn": "<http://www.wikidata.org/prop/direct-normalized/>",
            "wdv": "<http://www.wikidata.org/value/>",
            "wikibase": "<http://wikiba.se/ontology#>",
            "xsd": "<http://www.w3.org/2001/XMLSchema#>",
        }
        # see also https://www.wikidata.org/wiki/EntitySchema:E49
        sparql = ""
        for prefix in prefixes:
            if prefix in prefixMap:
                sparql += f"PREFIX {prefix}: {prefixMap[prefix]}\n"
        return sparql



"""
Created on 2022-07-24

@author: wf
"""
import json
import os
import urllib.parse
import urllib.request
from typing import List, Tuple


class WikidataSearch(object):
    """
    Wikidata Search API wrapper
    """

    def __init__(self, language: str = "en", timeout: float = 2.0):
        """
        Constructor

        Args:
            language(str): the language to use e.g. en/fr
            timeout(float): maximum time to wait for result
        """
        self.language = language
        self.timeout = timeout

    def searchOptions(
        self, searchFor: str, limit: int = 9
    ) -> List[Tuple[str, str, str]]:
        """
        Search and return a list of qid, itemLabel, description tuples.

        Args:
            searchFor (str): the string to search for.
            limit (int): the maximum amount of results to return.

        Returns:
            List[Tuple[str, str, str]]:
            A list of tuples containing
            qid, itemLabel, and description.
        """
        options = []
        srlist = self.search(searchFor, limit)
        if srlist is not None:
            for sr in srlist:
                qid = sr["id"]
                itemLabel = sr["label"]
                desc = ""
                if "display" in sr:
                    display = sr["display"]
                    if "description" in display:
                        desc = display["description"]["value"]
                options.append(
                    (
                        qid,
                        itemLabel,
                        desc,
                    )
                )
        return options

    def search(self, searchFor: str, limit: int = 9):
        """

        Args:
            searchFor(str): the string to search for
            limit(int): the maximum amount of results to search for
        """
        try:
            apiurl = f"https://www.wikidata.org/w/api.php?action=wbsearchentities&language={self.language}&uselang={self.language}&format=json&limit={limit}&search="
            searchEncoded = urllib.parse.quote_plus(searchFor)
            apisearch = apiurl + searchEncoded
            with urllib.request.urlopen(apisearch, timeout=self.timeout) as url:
                searchResult = json.loads(url.read().decode())
            return searchResult["search"]
        except Exception as _error:
            return None

    def getProperties(self):
        """
        get the Wikidata Properties
        """
        scriptdir = os.path.dirname(__file__)
        jsonPath = f"{scriptdir}/resources/wdprops.json"
        with open(jsonPath) as jsonFile:
            props = json.load(jsonFile)
        return props
